@@ -15,6 +15,8 @@
 #include <string.h>
 
 #include "soup-logger.h"
+#include "soup-connection.h"
+#include "soup-message-private.h"
 #include "soup.h"
 
 /**
@@ -67,7 +69,7 @@
  *
  * Currently, the request half of the message is logged just before
  * the first byte of the request gets written to the network (from the
- * #SoupSession::request_started signal), which means that if you have
+ * #SoupMessage::starting signal), which means that if you have
  * not made the complete request body available at that point, it will
  * not be logged.
  *
@@ -86,12 +88,6 @@
  * event of the #SoupMessage::finished signal.
  **/
 
-static void soup_logger_session_feature_init (SoupSessionFeatureInterface *feature_interface, gpointer interface_data);
-
-G_DEFINE_TYPE_WITH_CODE (SoupLogger, soup_logger, G_TYPE_OBJECT,
-			 G_IMPLEMENT_INTERFACE (SOUP_TYPE_SESSION_FEATURE,
-						soup_logger_session_feature_init))
-
 typedef struct {
 	/* We use a mutex so that if requests are being run in
 	 * multiple threads, we don't mix up the output.
@@ -101,6 +97,7 @@ typedef struct {
 	GQuark              tag;
 	GHashTable         *ids;
 
+	SoupSession        *session;
 	SoupLoggerLogLevel  level;
 	int                 max_body_size;
 
@@ -116,12 +113,28 @@ typedef struct {
 	gpointer            printer_data;
 	GDestroyNotify      printer_dnotify;
 } SoupLoggerPrivate;
-#define SOUP_LOGGER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SOUP_TYPE_LOGGER, SoupLoggerPrivate))
+
+enum {
+	PROP_0,
+
+	PROP_LEVEL,
+	PROP_MAX_BODY_SIZE,
+
+	LAST_PROP
+};
+
+static SoupSessionFeatureInterface *soup_logger_default_feature_interface;
+static void soup_logger_session_feature_init (SoupSessionFeatureInterface *feature_interface, gpointer interface_data);
+
+G_DEFINE_TYPE_WITH_CODE (SoupLogger, soup_logger, G_TYPE_OBJECT,
+                         G_ADD_PRIVATE (SoupLogger)
+			 G_IMPLEMENT_INTERFACE (SOUP_TYPE_SESSION_FEATURE,
+						soup_logger_session_feature_init))
 
 static void
 soup_logger_init (SoupLogger *logger)
 {
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 
 	g_mutex_init (&priv->lock);
 	priv->tag = g_quark_from_static_string (g_strdup_printf ("SoupLogger-%p", logger));
@@ -131,7 +144,8 @@ soup_logger_init (SoupLogger *logger)
 static void
 soup_logger_finalize (GObject *object)
 {
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (object);
+	SoupLogger *logger = SOUP_LOGGER (object);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 
 	g_hash_table_destroy (priv->ids);
 
@@ -147,14 +161,104 @@ soup_logger_finalize (GObject *object)
 	G_OBJECT_CLASS (soup_logger_parent_class)->finalize (object);
 }
 
+static void
+soup_logger_set_property (GObject *object, guint prop_id,
+			  const GValue *value, GParamSpec *pspec)
+{
+	SoupLogger *logger = SOUP_LOGGER (object);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
+
+	switch (prop_id) {
+	case PROP_LEVEL:
+		priv->level = g_value_get_enum (value);
+		break;
+	case PROP_MAX_BODY_SIZE:
+		priv->max_body_size = g_value_get_int (value);
+		break;
+	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+		break;
+	}
+}
+
+static void
+soup_logger_get_property (GObject *object, guint prop_id,
+			   GValue *value, GParamSpec *pspec)
+{
+	SoupLogger *logger = SOUP_LOGGER (object);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
+
+	switch (prop_id) {
+	case PROP_LEVEL:
+		g_value_set_enum (value, priv->level);
+		break;
+	case PROP_MAX_BODY_SIZE:
+		g_value_set_int (value, priv->max_body_size);
+		break;
+	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+		break;
+	}
+}
+
 static void
 soup_logger_class_init (SoupLoggerClass *logger_class)
 {
 	GObjectClass *object_class = G_OBJECT_CLASS (logger_class);
 
-	g_type_class_add_private (logger_class, sizeof (SoupLoggerPrivate));
-
 	object_class->finalize = soup_logger_finalize;
+	object_class->set_property = soup_logger_set_property;
+	object_class->get_property = soup_logger_get_property;
+
+	/* properties */
+	/**
+	 * SoupLogger:level:
+	 *
+	 * The level of logging output
+	 *
+	 * Since: 2.56
+	 */
+	/**
+	 * SOUP_LOGGER_LEVEL:
+	 *
+	 * Alias for the #SoupLogger:level property, qv.
+	 *
+	 * Since: 2.56
+	 **/
+	g_object_class_install_property (
+		object_class, PROP_LEVEL,
+		g_param_spec_enum (SOUP_LOGGER_LEVEL,
+				    "Level",
+				    "The level of logging output",
+				    SOUP_TYPE_LOGGER_LOG_LEVEL,
+				    SOUP_LOGGER_LOG_MINIMAL,
+				    G_PARAM_READWRITE));
+
+	/**
+	 * SoupLogger:max-body-size:
+	 *
+	 * If #SoupLogger:level is %SOUP_LOGGER_LOG_BODY, this gives
+	 * the maximum number of bytes of the body that will be logged.
+	 * (-1 means "no limit".)
+	 *
+	 * Since: 2.56
+	 */
+	/**
+	 * SOUP_LOGGER_MAX_BODY_SIZE:
+	 *
+	 * Alias for the #SoupLogger:max-body-size property, qv.
+	 *
+	 * Since: 2.56
+	 **/
+	g_object_class_install_property (
+		object_class, PROP_MAX_BODY_SIZE,
+		g_param_spec_int (SOUP_LOGGER_MAX_BODY_SIZE,
+				    "Max Body Size",
+				    "The maximum body size to output",
+				    -1,
+				    G_MAXINT,
+				    -1,
+				    G_PARAM_READWRITE));
 }
 
 /**
@@ -187,16 +291,10 @@ soup_logger_class_init (SoupLoggerClass *logger_class)
 SoupLogger *
 soup_logger_new (SoupLoggerLogLevel level, int max_body_size) 
 {
-	SoupLogger *logger;
-	SoupLoggerPrivate *priv;
-
-	logger = g_object_new (SOUP_TYPE_LOGGER, NULL);
-
-	priv = SOUP_LOGGER_GET_PRIVATE (logger);
-	priv->level = level;
-	priv->max_body_size = max_body_size;
-
-	return logger;
+	return g_object_new (SOUP_TYPE_LOGGER,
+			     SOUP_LOGGER_LEVEL, level,
+			     SOUP_LOGGER_MAX_BODY_SIZE, max_body_size,
+			     NULL);
 }
 
 /**
@@ -235,7 +333,7 @@ soup_logger_set_request_filter (SoupLogger       *logger,
 				gpointer          filter_data,
 				GDestroyNotify    destroy)
 {
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 
 	priv->request_filter         = request_filter;
 	priv->request_filter_data    = filter_data;
@@ -261,7 +359,7 @@ soup_logger_set_response_filter (SoupLogger       *logger,
 				 gpointer          filter_data,
 				 GDestroyNotify    destroy)
 {
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 
 	priv->response_filter         = response_filter;
 	priv->response_filter_data    = filter_data;
@@ -307,7 +405,7 @@ soup_logger_set_printer (SoupLogger        *logger,
 			 gpointer           printer_data,
 			 GDestroyNotify     destroy)
 {
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 
 	priv->printer         = printer;
 	priv->printer_data    = printer_data;
@@ -317,7 +415,7 @@ soup_logger_set_printer (SoupLogger        *logger,
 static guint
 soup_logger_get_id (SoupLogger *logger, gpointer object)
 {
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 
 	return GPOINTER_TO_UINT (g_object_get_qdata (object, priv->tag));
 }
@@ -325,7 +423,7 @@ soup_logger_get_id (SoupLogger *logger, gpointer object)
 static guint
 soup_logger_set_id (SoupLogger *logger, gpointer object)
 {
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 	gpointer klass = G_OBJECT_GET_CLASS (object);
 	gpointer id;
 
@@ -381,7 +479,7 @@ static void
 soup_logger_print (SoupLogger *logger, SoupLoggerLogLevel level,
 		   char direction, const char *format, ...)
 {
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 	va_list args;
 	char *data, *line, *end;
 
@@ -443,13 +541,13 @@ soup_logger_print_basic_auth (SoupLogger *logger, const char *value)
 
 static void
 print_request (SoupLogger *logger, SoupMessage *msg,
-	       SoupSession *session, SoupSocket *socket,
-	       gboolean restarted)
+	       SoupSocket *socket, gboolean restarted)
 {
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 	SoupLoggerLogLevel log_level;
 	SoupMessageHeadersIter iter;
 	const char *name, *value;
+	char *socket_dbg;
 	SoupURI *uri;
 
 	if (priv->request_filter) {
@@ -479,15 +577,22 @@ print_request (SoupLogger *logger, SoupMessage *msg,
 	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, '>',
 			   "Soup-Debug-Timestamp: %lu",
 			   (unsigned long)time (0));
+
+	socket_dbg = socket ?
+		g_strdup_printf ("%s %u (%p)",
+				 g_type_name_from_instance ((GTypeInstance *)socket),
+				 soup_logger_get_id (logger, socket), socket)
+		: NULL;
+
 	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, '>',
-			   "Soup-Debug: %s %u (%p), %s %u (%p), %s %u (%p)%s",
-			   g_type_name_from_instance ((GTypeInstance *)session),
-			   soup_logger_get_id (logger, session), session,
+			   "Soup-Debug: %s %u (%p), %s %u (%p), %s%s",
+			   g_type_name_from_instance ((GTypeInstance *)priv->session),
+			   soup_logger_get_id (logger, priv->session), priv->session,
 			   g_type_name_from_instance ((GTypeInstance *)msg),
 			   soup_logger_get_id (logger, msg), msg,
-			   g_type_name_from_instance ((GTypeInstance *)socket),
-			   soup_logger_get_id (logger, socket), socket,
+			   socket_dbg ? socket_dbg : "cached",
 			   restarted ? ", restarted" : "");
+	g_free (socket_dbg);
 
 	if (log_level == SOUP_LOGGER_LOG_MINIMAL)
 		return;
@@ -541,7 +646,7 @@ print_request (SoupLogger *logger, SoupMessage *msg,
 static void
 print_response (SoupLogger *logger, SoupMessage *msg)
 {
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 	SoupLoggerLogLevel log_level;
 	SoupMessageHeadersIter iter;
 	const char *name, *value;
@@ -589,7 +694,7 @@ static void
 finished (SoupMessage *msg, gpointer user_data)
 {
 	SoupLogger *logger = user_data;
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 
 	g_mutex_lock (&priv->lock);
 
@@ -603,7 +708,7 @@ static void
 got_informational (SoupMessage *msg, gpointer user_data)
 {
 	SoupLogger *logger = user_data;
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 
 	g_mutex_lock (&priv->lock);
 
@@ -638,7 +743,7 @@ static void
 got_body (SoupMessage *msg, gpointer user_data)
 {
 	SoupLogger *logger = user_data;
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 
 	g_mutex_lock (&priv->lock);
 
@@ -649,6 +754,36 @@ got_body (SoupMessage *msg, gpointer user_data)
 	g_mutex_unlock (&priv->lock);
 }
 
+static void
+starting (SoupMessage *msg, gpointer user_data)
+{
+	SoupLogger *logger = SOUP_LOGGER (user_data);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
+	gboolean restarted;
+	guint msg_id;
+	SoupConnection *conn;
+	SoupSocket *socket;
+
+	msg_id = soup_logger_get_id (logger, msg);
+	if (msg_id)
+		restarted = TRUE;
+	else {
+		soup_logger_set_id (logger, msg);
+		restarted = FALSE;
+	}
+
+	if (!soup_logger_get_id (logger, priv->session))
+		soup_logger_set_id (logger, priv->session);
+
+	conn = soup_message_get_connection (msg);
+	socket = conn ? soup_connection_get_socket (conn) : NULL;
+	if (socket && !soup_logger_get_id (logger, socket))
+		soup_logger_set_id (logger, socket);
+
+	print_request (logger, msg, socket, restarted);
+	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, ' ', "");
+}
+
 static void
 soup_logger_request_queued (SoupSessionFeature *logger,
 			    SoupSession *session,
@@ -656,6 +791,9 @@ soup_logger_request_queued (SoupSessionFeature *logger,
 {
 	g_return_if_fail (SOUP_IS_MESSAGE (msg));
 
+	g_signal_connect (msg, "starting",
+			  G_CALLBACK (starting),
+			  logger);
 	g_signal_connect (msg, "got-informational",
 			  G_CALLBACK (got_informational),
 			  logger);
@@ -667,38 +805,6 @@ soup_logger_request_queued (SoupSessionFeature *logger,
 			  logger);
 }
 
-static void
-soup_logger_request_started (SoupSessionFeature *feature,
-			     SoupSession *session,
-			     SoupMessage *msg,
-			     SoupSocket *socket)
-{
-	SoupLogger *logger = SOUP_LOGGER (feature);
-	gboolean restarted;
-	guint msg_id;
-
-	g_return_if_fail (SOUP_IS_SESSION (session));
-	g_return_if_fail (SOUP_IS_MESSAGE (msg));
-	g_return_if_fail (SOUP_IS_SOCKET (socket));
-
-	msg_id = soup_logger_get_id (logger, msg);
-	if (msg_id)
-		restarted = TRUE;
-	else {
-		soup_logger_set_id (logger, msg);
-		restarted = FALSE;
-	}
-
-	if (!soup_logger_get_id (logger, session))
-		soup_logger_set_id (logger, session);
-
-	if (!soup_logger_get_id (logger, socket))
-		soup_logger_set_id (logger, socket);
-
-	print_request (logger, msg, session, socket, restarted);
-	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, ' ', "");
-}
-
 static void
 soup_logger_request_unqueued (SoupSessionFeature *logger,
 			      SoupSession *session,
@@ -706,16 +812,31 @@ soup_logger_request_unqueued (SoupSessionFeature *logger,
 {
 	g_return_if_fail (SOUP_IS_MESSAGE (msg));
 
+	g_signal_handlers_disconnect_by_func (msg, starting, logger);
 	g_signal_handlers_disconnect_by_func (msg, got_informational, logger);
 	g_signal_handlers_disconnect_by_func (msg, got_body, logger);
 	g_signal_handlers_disconnect_by_func (msg, finished, logger);
 }
 
+static void
+soup_logger_feature_attach (SoupSessionFeature *feature,
+			    SoupSession *session)
+{
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (SOUP_LOGGER (feature));
+
+	priv->session = session;
+
+	soup_logger_default_feature_interface->attach (feature, session);
+}
+
 static void
 soup_logger_session_feature_init (SoupSessionFeatureInterface *feature_interface,
 				  gpointer interface_data)
 {
+	soup_logger_default_feature_interface =
+		g_type_default_interface_peek (SOUP_TYPE_SESSION_FEATURE);
+
+	feature_interface->attach = soup_logger_feature_attach;
 	feature_interface->request_queued = soup_logger_request_queued;
-	feature_interface->request_started = soup_logger_request_started;
 	feature_interface->request_unqueued = soup_logger_request_unqueued;
 }
